  /********************************************************************
   *                                                                  *
   *  Project: 16F1828 4-Digit Photo Timer                            *
   *   Source: 16F1828_4-Digit_Photo_Timer.c                          *
   *   Author: Mike McLaren, K8LH                                     *
   *  (C)2012: Micro Application Consultants                          *
   *     Date: 29-Apr-2012                                            *
   *                                                                  *
   *   16F1828 "4-Digit 99 Minute Timer" Experiment (lighting one     *
   *   segment at a time at 1/32nd duty cycle).                       *
   *                                                                  *
   *      IDE: MPLAB 8.84 (tabs = 4)                                  *
   *     Lang: SourceBoost BoostC v7.05, Lite/Free version            *
   *                                                                  *
   ********************************************************************/

   #include <system.h>

   #pragma DATA _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _MCLRE_OFF
   #pragma DATA _CONFIG2, _PLLEN_OFF & _LVP_OFF

   #pragma CLOCK_FREQ 8000000       // 8 MHz INTOSC

  /********************************************************************
   *  function prototypes                                             *
   ********************************************************************/

  /********************************************************************
   *  typedef and defines                                             *
   ********************************************************************/

   typedef unsigned char u08;
   typedef unsigned int u16;

   #define r08 const rom unsigned char

   #define timeout (seconds | minutes) == 0

   #define relay porta.2            // active hi relay output (RA2)
   #define piezo porta.1            // piezo speaker (RA1)

  /********************************************************************
   *  variables (and const arrays)                                    *
   ********************************************************************/

   u08 colsel = 0x10;               // column select mask
   u08 segmask = 0x01;              // segment select mask
   u08 segdata;                     //
   u08 group = 0;                   // set mode flash group mask
   u08 flash = 0;                   // set mode flash timer

   u08 seconds = 0x00;              // packed bcd seconds
   u08 minutes = 0x00;              // packed bcd minutes
   u16 isrtimer = 4000;             // isr 1 second timer

   u08 swnew = 0;                   // switch state logic work var
   u08 swold = 0;                   // switch state latch
   u08 flags = 0;                   // switch flags (bits 7..4)
   u08 beepctr = 0;                 // beep task var

   r08 segment[] = { 0b00111111,    // "0"   -|-|F|E|D|C|B|A
                     0b00000110,    // "1"   -|-|-|-|-|C|B|-
                     0b01011011,    // "2"   -|G|-|E|D|-|B|A
                     0b01001111,    // "3"   -|G|-|-|D|C|B|A
                     0b01100110,    // "4"   -|G|F|-|-|C|B|-
                     0b01101101,    // "5"   -|G|F|-|D|C|-|A
                     0b01111101,    // "6"   -|G|F|E|D|C|-|A
                     0b00000111,    // "7"   -|-|-|-|-|C|B|A
                     0b01111111,    // "8"   -|G|F|E|D|C|B|A
                     0b01101111,    // "9"   -|G|F|-|D|C|B|A
                     0b00000000,    // " "   -|-|-|-|-|-|-|-
                     0b01000000 };  // "-"   -|G|-|-|-|-|-|-


  /********************************************************************
   *  functions                                                       *
   ********************************************************************/


  /********************************************************************
   *  main init                                                       *
   ********************************************************************/

   void main()
   {
     ansela = 0;                //
     anselb = 0;                //
     anselc = 0;                //
     porta = 0;                 //
     portb = 0;                 //
     portc = 0;                 //
     trisa = 0;                 //
     trisb = 0;                 //
     trisc = 0;                 //
     osccon = 0b01110000;       // set INTOSC to 8 MHz
     while(!oscstat.HFIOFS);    // wait for oscillator stable
  /*                                                                  *
   *  250-us (500 cycle) TMR2 interrupts for a 125-Hz refresh rate    *
   *                                                                  */
     tmr2 = 0;                  // clear Timer 2 register
     t2con = 0b00000001;        // '0-------' unimplemented bit
                                // '-0000---' TOUTPS<3:0>, post 1
                                // '-----0--' TMR2ON, Timer 2 off
                                // '------01' T2CKPS<1:0>, pre 4
     pr2 = 125-1;               // 125 x 2-usec 'ticks' = 250 usecs
     pir1 = 0;                  // clear peripheral interrupt flags
     pie1.TMR2IE = 1;           // set Timer 2 interrupt enable bit
     intcon = 0b11000000;       // '1-------' GIE, enable global and
                                // '-1------' PEIE, peripheral ints
                                // '--0-----' T0IE, TMR0 ints off
                                // '---0----' INTE, off
                                // '----0---' GPIE, IOC disabled
                                // '-----000' T0IF/INTF/GPIF flags
     t2con.TMR2ON = 1;          // start TMR2 (and interrupts)
     
  /********************************************************************
   *  main loop                                                       *
   ********************************************************************/

     while(1)                   //
     { 
      /****************************************************************
       *  run mode                                                    *
       ****************************************************************/
       if(flags.7)              // if rt arrow switch
       { flags.7 = 0;           // clear switch flag and
         if(minutes | seconds)  // if not timed out then
         { isrtimer = 4000;     // reset ISR 1-second timer
           relay ^= 1;          // toggle the relay output
         }                      //
       }                        //

      /****************************************************************
       *  set mode                                                    *
       ****************************************************************/
       if(flags.4)              // if set switch "on"
       { 
         if(relay == 0)         // if timer not running
           group = 0b00110000;  // flash "minutes" display group
         else                   // else (timer running)
           flags.4 = 0;         // cancel "set" mode

         while(flags.4)         // while in set mode
         { 
          /*                                                          *
           *  rt arrow toggles between "minutes" & "seconds" groups   *
           *                                                          */
           if(flags.7)          // if rt arrow
           { flags.7 = 0;       // clear switch flag and
             group ^= 0xF0;     // toggle flash group
           }                    //

          /*                                                          *
           *  up arrow increments the packed bcd group value          *
           *                                                          */
           if(swold.5)          // if up arrow
           { if(group.4)        // if "minutes" group
               if(minutes++ == 0x99) minutes = 0x00;
             if((minutes & 15) == 10) minutes += 6;

             if(group.6)        // if "seconds" group
               if(seconds++ == 0x59) seconds = 0x00;
             if((seconds & 15) == 10) seconds += 6;
             delay_ms(200);     // repeat interval
           }                    //

          /*                                                          *
           *  dn arrow decrements the packed bcd group value          *
           *                                                          */
           if(swold.6)          // if dn arrow
           { if(group.4)        // if "minutes" group
               if(minutes-- == 0x00) minutes = 0x99;
             if((minutes & 15) == 15) minutes -= 6;

             if(group.6)        // if "seconds" group
               if(seconds-- == 0x00) seconds = 0x59;
             if((seconds & 15) == 15) seconds -= 6;
             delay_ms(200);     // repeat interval
           }                    //
         }                      //
         group = 0;             // flash off
         flags = 0;             // clear switch flags
       }                        //
     }
   }

  /********************************************************************
   *  interrupt                                                       *
   ********************************************************************/

   void interrupt()             // 250-us (500 cyc) TMR2 interrupts
   { u08 index;                 //
     pir1.TMR2IF = 0;           // clear TMR2 interrupt flag
    /*                                                                *
     *  refresh display (display new segment)                         *
     *                                                                */
     portb = 0;                 // blank the display
     portc = ~segdata;          // new segment (invert, active lo)
     portb = colsel;            // light new segment (column "on")

    /*                                                                *
     *  beep task generates a 32-ms 2000-Hz "new press" beep          *
     *                                                                */
     if(beepctr)                // if beep task running
     { piezo ^= 1;              // toggle piezo spkr pin 
       beepctr--;               //
     }                          //

    /*                                                                *
     *  K8LH Parallel Switch State Logic (8 ms debounce intervals)    *
     *                                                                *
     *  swnew  ___---___---______  sample active hi switches          *
     *  swold  ____---___---_____  switch state latch                 *
     *  swnew  ___-__-__-__-_____  changes, press or release          *
     *  swnew  ___-_____-________  filter out 'release' bits          *
     *  flags  ___-------________  toggle flag bits for main          *
     *                                                                */
     if(segmask.0)              // if segment A (2-ms) interval
     { swnew = swold | colsel;  // indicate pressed unconditionally
       if(porta.3 == 0)         // if not pressed (RA3 = 0)
         swnew ^= colsel;       // indicate released
       swnew ^= swold;          // changes, press or release
       swold ^= swnew;          // update switch state latch
       swnew &= swold;          // filter out 'release' bits
       if(swnew)                // if any "new press" bits
         beepctr = 128;         // task a "new press" beep
       flags ^= swnew;          // toggle flag bits for main
     }                          //

    /*                                                                *
     *  advance segment mask bit for next interrupt and if last       *
     *  segment, bump column select bit mask.                         *
     *                                                                */
     segmask <<= 1;             // 
     if(segmask == 0)           // if last segment (2-ms interval)
     { segmask = 1;             // reset to segment A and
       colsel <<= 1;            // advance column select bit
       if(colsel == 0)          // if last column
         colsel = 0x10;         // reset to first column (RB4)
       flash++;                 // bump 'flash' timer
     }                          //

    /*                                                                *
     *  prepare new segment data for next interrupt                   *
     *                                                                */
     index = minutes >> 4;      // 0..9
     if(!index) index = 10;     // suppress a leading zero
     if(colsel.5) index = minutes & 15;
     if(colsel.6) index = seconds >> 4;
     if(colsel.7) index = seconds & 15;
     segdata = segment[index];  //
     if(colsel.4)               // if column/digit 1
       segdata.7 = 1;           // colon on (decimal pt segment)
     segdata &= segmask;        // only one segment each interrupt

    /*                                                                *
     *  flash MINS or SECS display group at 2-Hz rate in set mode     *
     *                                                                */
     if(flags.4)                // if SET switch "on" and
     { if(colsel & group)       // if SET display group and
       { if(flash.7)            // if 256-ms flash interval
           segdata &= 0x80;     // turn the segment off
       }                        //
     }                          //

    /*                                                                *
     *  count-down timer function                                     *
     *                                                                */
     if(relay)                  // if count down timer running
     { if(--isrtimer == 0)      // if 1 second interval
       { isrtimer = 4000;       // reset 1 second timer and
         if(seconds-- == 0)     // decrement count down timer
         { seconds = 0x59;      //
           minutes--;           //
         }                      //
       }                        //
       if((minutes & 15) == 15) // packed bcd adjust
         minutes -= 6;          //
       if((seconds & 15) == 15) // packed bcd adjust
         seconds -= 6;          //
       if(timeout)              // if timer timed-out
       { relay = 0;             // turn relay output off
         beepctr.7 = 1;         // task a "beep"
       }                        //
     }                          //
   }                            //

